#! /usr/bin/python

from twisted.trial import unittest

from buildbot.changes import pb

d1 = {'files': ["Project/foo.c", "Project/bar/boo.c"],
      'who': "marvin",
      'comments': "Some changes in Project"}
d2 = {'files': ["OtherProject/bar.c"],
      'who': "zaphod",
      'comments': "other changes"}
d3 = {'files': ["Project/baz.c", "OtherProject/bloo.c"],
      'who': "alice",
      'comments': "mixed changes"}

class TestChangePerspective(unittest.TestCase):

    def setUp(self):
        self.changes = []

    def addChange(self, c):
        self.changes.append(c)

    def testNoPrefix(self):
        p = pb.ChangePerspective(self, None)
        p.perspective_addChange(d1)
        self.failUnlessEqual(len(self.changes), 1)
        c1 = self.changes[0]
        self.failUnlessEqual(c1.files,
                             ["Project/foo.c", "Project/bar/boo.c"])
        self.failUnlessEqual(c1.comments, "Some changes in Project")
        self.failUnlessEqual(c1.who, "marvin")

    def testPrefix(self):
        p = pb.ChangePerspective(self, "Project")

        p.perspective_addChange(d1)
        self.failUnlessEqual(len(self.changes), 1)
        c1 = self.changes[-1]
        self.failUnlessEqual(c1.files, ["foo.c", "bar/boo.c"])
        self.failUnlessEqual(c1.comments, "Some changes in Project")
        self.failUnlessEqual(c1.who, "marvin")

        p.perspective_addChange(d2) # should be ignored
        self.failUnlessEqual(len(self.changes), 1)

        p.perspective_addChange(d3) # should ignore the OtherProject file
        self.failUnlessEqual(len(self.changes), 2)

        c3 = self.changes[-1]
        self.failUnlessEqual(c3.files, ["baz.c"])
        self.failUnlessEqual(c3.comments, "mixed changes")
        self.failUnlessEqual(c3.who, "alice")

