#! /usr/bin/python

from __future__ import generators
import os, os.path

from twisted.trial import unittest
from twisted.python import components

try:
    import cvstoys
    from buildbot.changes.freshcvs import FreshCVSSource
except ImportError:
    cvstoys = None

from buildbot.master import BuildMaster
from twisted.application import service, internet
from twisted.spread import pb
from twisted.web.server import Site
from twisted.web.distrib import ResourcePublisher
from buildbot.process.builder import Builder
from buildbot.process.factory import BasicBuildFactory
from buildbot.process.interlock import Interlock
from buildbot.process import step
from buildbot.status import words, html, builder

import sys
from twisted.python import log
#log.startLogging(sys.stdout)

emptyCfg = \
"""
c = {}
c['bots'] = []
c['sources'] = []
c['builders'] = []
c['slavePortnum'] = 9999
c['projectName'] = 'dummy project'
c['projectURL'] = 'http://dummy.example.com'
c['buildbotURL'] = 'http://dummy.example.com/buildbot'
BuildmasterConfig = c
"""

slaveportCfg = \
"""
c = {}
c['bots'] = []
c['sources'] = []
c['builders'] = []
c['slavePortnum'] = 9000
BuildmasterConfig = c
"""

botsCfg = \
"""
c = {}
c['bots'] = [('bot1', 'pw1'), ('bot2', 'pw2')]
c['sources'] = []
c['builders'] = []
c['slavePortnum'] = 9999
BuildmasterConfig = c
"""

sourcesCfg = \
"""
from buildbot.changes.freshcvs import FreshCVSSource
c = {}
c['bots'] = []
s1 = FreshCVSSource('cvs.example.com', 1000, 'pname', 'spass',
                    prefix='Prefix/')
c['sources'] = [s1]
c['builders'] = []
c['slavePortnum'] = 9999
BuildmasterConfig = c
"""

buildersCfg = \
"""
from buildbot.process.factory import BasicBuildFactory
c = {}
c['bots'] = [('bot1', 'pw1')]
c['sources'] = []
f1 = BasicBuildFactory('cvsroot', 'cvsmodule')
c['builders'] = [('builder1', 'bot1', 'workdir', f1)]
c['slavePortnum'] = 9999
BuildmasterConfig = c
"""

buildersCfg2 = \
"""
from buildbot.process.factory import BasicBuildFactory
c = {}
c['bots'] = [('bot1', 'pw1')]
c['sources'] = []
f1 = BasicBuildFactory('cvsroot', 'cvsmodule2')
c['builders'] = [('builder1', 'bot1', 'workdir', f1)]
c['slavePortnum'] = 9999
BuildmasterConfig = c
"""

buildersCfg2new = \
"""
from buildbot.process.factory import BasicBuildFactory
c = {}
c['bots'] = [('bot1', 'pw1')]
c['sources'] = []
f1 = BasicBuildFactory('cvsroot', 'cvsmodule2')
c['builders'] = [{ 'name': 'builder1', 'slavename': 'bot1',
                   'builddir': 'workdir', 'factory': f1 }]
c['slavePortnum'] = 9999
BuildmasterConfig = c
"""

buildersCfg1new = \
"""
from buildbot.process.factory import BasicBuildFactory
c = {}
c['bots'] = [('bot1', 'pw1')]
c['sources'] = []
f1 = BasicBuildFactory('cvsroot', 'cvsmodule')
c['builders'] = [{ 'name': 'builder1', 'slavename': 'bot1',
                   'builddir': 'workdir', 'factory': f1 }]
c['slavePortnum'] = 9999
BuildmasterConfig = c
"""

buildersCfg3 = \
"""
from buildbot.process.factory import BasicBuildFactory
c = {}
c['bots'] = [('bot1', 'pw1')]
c['sources'] = []
f1 = BasicBuildFactory('cvsroot', 'cvsmodule2')
c['builders'] = [{ 'name': 'builder1', 'slavename': 'bot1',
                   'builddir': 'workdir', 'factory': f1 },
                 { 'name': 'builder2', 'slavename': 'bot1',
                   'builddir': 'workdir2', 'factory': f1 }]
c['slavePortnum'] = 9999
BuildmasterConfig = c
"""

buildersCfg4 = \
"""
from buildbot.process.factory import BasicBuildFactory
c = {}
c['bots'] = [('bot1', 'pw1')]
c['sources'] = []
f1 = BasicBuildFactory('cvsroot', 'cvsmodule2')
c['builders'] = [{ 'name': 'builder1', 'slavename': 'bot1',
                   'builddir': 'newworkdir', 'factory': f1 },
                 { 'name': 'builder2', 'slavename': 'bot1',
                   'builddir': 'workdir2', 'factory': f1 }]
c['slavePortnum'] = 9999
BuildmasterConfig = c
"""

ircCfg1 = \
"""
c = {}
c['bots'] = []
c['sources'] = []
c['builders'] = []
c['slavePortnum'] = 9999
c['irc'] = {('irc.us.freenode.net', 6667): ('buildbot', ['twisted'])}
BuildmasterConfig = c
"""

ircCfg2 = \
"""
from buildbot.status import words
c = {}
c['bots'] = []
c['sources'] = []
c['builders'] = []
c['slavePortnum'] = 9999
c['status'] = [words.IRC('irc.us.freenode.net', 'buildbot', ['twisted'])]
BuildmasterConfig = c
"""

ircCfg3 = \
"""
from buildbot.status import words
c = {}
c['bots'] = []
c['sources'] = []
c['builders'] = []
c['slavePortnum'] = 9999
c['status'] = [words.IRC('irc.us.freenode.net', 'buildbot', ['twisted']),
               words.IRC('irc.example.com', 'otherbot', ['chan1', 'chan2'])]
BuildmasterConfig = c
"""

ircCfg4 = \
"""
from buildbot.status import words
c = {}
c['bots'] = []
c['sources'] = []
c['builders'] = []
c['slavePortnum'] = 9999
c['status'] = [words.IRC('irc.us.freenode.net', 'buildbot', ['knotted'])]
BuildmasterConfig = c
"""

webCfg1 = \
"""
c = {}
c['bots'] = []
c['sources'] = []
c['builders'] = []
c['slavePortnum'] = 9999
c['webPortnum'] = 9980  # backwards compatibility
BuildmasterConfig = c
"""

webCfg2 = \
"""
from buildbot.status import html
c = {}
c['bots'] = []
c['sources'] = []
c['builders'] = []
c['slavePortnum'] = 9999
c['status'] = [html.Waterfall(http_port=9980)]
BuildmasterConfig = c
"""

webCfg3 = \
"""
from buildbot.status import html
c = {}
c['bots'] = []
c['sources'] = []
c['builders'] = []
c['slavePortnum'] = 9999
c['status'] = [html.Waterfall(http_port=9981)]
BuildmasterConfig = c
"""

webNameCfg1 = \
"""
c = {}
c['bots'] = []
c['sources'] = []
c['builders'] = []
c['slavePortnum'] = 9999
c['webPathname'] = '~/.twistd-web-pb' # backwards compatibility
BuildmasterConfig = c
"""

webNameCfg2 = \
"""
from buildbot.status import html
c = {}
c['bots'] = []
c['sources'] = []
c['builders'] = []
c['slavePortnum'] = 9999
c['status'] = [html.Waterfall(distrib_port='~/.twistd-web-pb')]
BuildmasterConfig = c
"""

webNameCfg3 = \
"""
from buildbot.status import html
c = {}
c['bots'] = []
c['sources'] = []
c['builders'] = []
c['slavePortnum'] = 9999
c['status'] = [html.Waterfall(distrib_port='bar.socket')]
BuildmasterConfig = c
"""

debugPasswordCfg = \
"""
c = {}
c['bots'] = []
c['sources'] = []
c['builders'] = []
c['slavePortnum'] = 9999
c['debugPassword'] = 'sekrit'
BuildmasterConfig = c
"""

# create an inactive interlock (builder3 is not yet defined). This isn't
# recommended practice, it is only here to test the code
interlockCfg1 = \
"""
from buildbot.process.factory import BasicBuildFactory
c = {}
c['bots'] = [('bot1', 'pw1')]
c['sources'] = []
f1 = BasicBuildFactory('cvsroot', 'cvsmodule')
c['builders'] = [('builder1', 'bot1', 'workdir', f1),
                 ('builder2', 'bot1', 'workdir2', f1),
                 ('builder4', 'bot1', 'workdir4', f1),
                 ('builder5', 'bot1', 'workdir5', f1),
                 ]
c['interlocks'] = [('lock1', ['builder1'], ['builder2', 'builder3']),
                   ]
c['slavePortnum'] = 9999
BuildmasterConfig = c
"""

# make it active
interlockCfg2 = \
"""
from buildbot.process.factory import BasicBuildFactory
c = {}
c['bots'] = [('bot1', 'pw1')]
c['sources'] = []
f1 = BasicBuildFactory('cvsroot', 'cvsmodule')
c['builders'] = [('builder1', 'bot1', 'workdir', f1),
                 ('builder2', 'bot1', 'workdir2', f1),
                 ('builder3', 'bot1', 'workdir3', f1),
                 ('builder4', 'bot1', 'workdir4', f1),
                 ('builder5', 'bot1', 'workdir5', f1),
                 ]
c['interlocks'] = [('lock1', ['builder1'], ['builder2', 'builder3']),
                   ]
c['slavePortnum'] = 9999
BuildmasterConfig = c
"""

# add a second lock
interlockCfg3 = \
"""
from buildbot.process.factory import BasicBuildFactory
c = {}
c['bots'] = [('bot1', 'pw1')]
c['sources'] = []
f1 = BasicBuildFactory('cvsroot', 'cvsmodule')
c['builders'] = [('builder1', 'bot1', 'workdir', f1),
                 ('builder2', 'bot1', 'workdir2', f1),
                 ('builder3', 'bot1', 'workdir3', f1),
                 ('builder4', 'bot1', 'workdir4', f1),
                 ('builder5', 'bot1', 'workdir5', f1),
                 ]
c['interlocks'] = [('lock1', ['builder1'], ['builder2', 'builder3']),
                   ('lock2', ['builder3', 'builder4'], ['builder5']),
                   ]
c['slavePortnum'] = 9999
BuildmasterConfig = c
"""

# change the second lock
interlockCfg4 = \
"""
from buildbot.process.factory import BasicBuildFactory
c = {}
c['bots'] = [('bot1', 'pw1')]
c['sources'] = []
f1 = BasicBuildFactory('cvsroot', 'cvsmodule')
c['builders'] = [('builder1', 'bot1', 'workdir', f1),
                 ('builder2', 'bot1', 'workdir2', f1),
                 ('builder3', 'bot1', 'workdir3', f1),
                 ('builder4', 'bot1', 'workdir4', f1),
                 ('builder5', 'bot1', 'workdir5', f1),
                 ]
c['interlocks'] = [('lock1', ['builder1'], ['builder2', 'builder3']),
                   ('lock2', ['builder1', 'builder4'], ['builder5']),
                   ]
c['slavePortnum'] = 9999
BuildmasterConfig = c
"""

# delete the first lock
interlockCfg5 = \
"""
from buildbot.process.factory import BasicBuildFactory
c = {}
c['bots'] = [('bot1', 'pw1')]
c['sources'] = []
f1 = BasicBuildFactory('cvsroot', 'cvsmodule')
c['builders'] = [('builder1', 'bot1', 'workdir', f1),
                 ('builder2', 'bot1', 'workdir2', f1),
                 ('builder3', 'bot1', 'workdir3', f1),
                 ('builder4', 'bot1', 'workdir4', f1),
                 ('builder5', 'bot1', 'workdir5', f1),
                 ]
c['interlocks'] = [('lock2', ['builder1', 'builder4'], ['builder5']),
                   ]
c['slavePortnum'] = 9999
BuildmasterConfig = c
"""

# render the lock inactive by removing a builder it depends upon
interlockCfg6 = \
"""
from buildbot.process.factory import BasicBuildFactory
c = {}
c['bots'] = [('bot1', 'pw1')]
c['sources'] = []
f1 = BasicBuildFactory('cvsroot', 'cvsmodule')
c['builders'] = [('builder1', 'bot1', 'workdir', f1),
                 ('builder2', 'bot1', 'workdir2', f1),
                 ('builder3', 'bot1', 'workdir3', f1),
                 ('builder4', 'bot1', 'workdir4', f1),
                 ]
c['interlocks'] = [('lock2', ['builder1', 'builder4'], ['builder5']),
                   ]
c['slavePortnum'] = 9999
BuildmasterConfig = c
"""

# finally remove the interlock
interlockCfg7 = \
"""
from buildbot.process.factory import BasicBuildFactory
c = {}
c['bots'] = [('bot1', 'pw1')]
c['sources'] = []
f1 = BasicBuildFactory('cvsroot', 'cvsmodule')
c['builders'] = [('builder1', 'bot1', 'workdir', f1),
                 ('builder2', 'bot1', 'workdir2', f1),
                 ('builder3', 'bot1', 'workdir3', f1),
                 ('builder4', 'bot1', 'workdir4', f1),
                 ]
c['interlocks'] = []
c['slavePortnum'] = 9999
BuildmasterConfig = c
"""

class ConfigTest(unittest.TestCase):
    def setUp(self):
        self.buildmaster = BuildMaster(".")
        
    def failUnlessListsEquivalent(self, list1, list2):
        l1 = list1[:]
        l1.sort()
        l2 = list2[:]
        l2.sort()
        self.failUnlessEqual(l1, l2)

    def servers(self, s, types):
        # perform a recursive search of s.services, looking for instances of
        # twisted.application.internet.TCPServer, then extract their .args
        # values to find the TCP ports they want to listen on
        for child in s:
            if components.implements(child, service.IServiceCollection):
                for gc in self.servers(child, types):
                    yield gc
            if isinstance(child, types):
                yield child

    def TCPports(self, s):
        return list(self.servers(s, internet.TCPServer))
    def UNIXports(self, s):
        return list(self.servers(s, internet.UNIXServer))
    def TCPclients(self, s):
        return list(self.servers(s, internet.TCPClient))

    def checkPorts(self, svc, expected):
        """Verify that the TCPServer and UNIXServer children of the given
        service have the expected portnum/pathname and factory classes. As a
        side-effect, return a list of servers in the same order as the
        'expected' list. This can be used to verify properties of the
        factories contained therein."""
        
        expTCP = [e for e in expected if type(e[0]) == int]
        expUNIX = [e for e in expected if type(e[0]) == str]
        haveTCP = [(p.args[0], p.args[1].__class__)
                   for p in self.TCPports(svc)]
        haveUNIX = [(p.args[0], p.args[1].__class__)
                    for p in self.UNIXports(svc)]
        self.failUnlessListsEquivalent(expTCP, haveTCP)
        self.failUnlessListsEquivalent(expUNIX, haveUNIX)
        ret = []
        for e in expected:
            for have in self.TCPports(svc) + self.UNIXports(svc):
                if have.args[0] == e[0]:
                    ret.append(have)
                    continue
        assert(len(ret) == len(expected))
        return ret

    def testEmpty(self):
        self.failUnlessRaises(KeyError, self.buildmaster.loadConfig, "")

    def testSimple(self):
        # covers slavePortnum, base checker passwords
        master = self.buildmaster
        master.loadChanges()

        master.loadConfig(emptyCfg)
        # note: this doesn't actually start listening, because the app
        # hasn't been started running
        self.failUnlessEqual(master.slavePortnum, 9999)
        self.checkPorts(master, [(9999, pb.PBServerFactory)])
        self.failUnlessEqual(master.sources, [])
        self.failUnlessEqual(master.change_svc.sources, [])
        self.failUnlessEqual(master.botmaster.builders, {})
        self.failUnlessEqual(master.botmaster.interlocks, {})
        self.failUnlessEqual(master.checker.users,
                             {"change": "changepw"})
        self.failUnlessEqual(master.projectName, "dummy project")
        self.failUnlessEqual(master.projectURL, "http://dummy.example.com")
        self.failUnlessEqual(master.buildbotURL,
                             "http://dummy.example.com/buildbot")

    def testSlavePortnum(self):
        master = self.buildmaster
        master.loadChanges()

        master.loadConfig(emptyCfg)
        self.failUnlessEqual(master.slavePortnum, 9999)
        ports = self.checkPorts(master, [(9999, pb.PBServerFactory)])
        p = ports[0]

        master.loadConfig(emptyCfg)
        self.failUnlessEqual(master.slavePortnum, 9999)
        ports = self.checkPorts(master, [(9999, pb.PBServerFactory)])
        self.failUnlessIdentical(p, ports[0],
                                 "the slave port was changed even " + \
                                 "though the configuration was not")

        master.loadConfig(slaveportCfg)
        self.failUnlessEqual(master.slavePortnum, 9000)
        ports = self.checkPorts(master, [(9000, pb.PBServerFactory)])
        self.failIf(p is ports[0],
                    "slave port was unchanged but configuration was changed")

    def testBots(self):
        master = self.buildmaster
        master.loadChanges()
        master.loadConfig(emptyCfg)
        self.failUnlessEqual(master.botmaster.builders, {})
        self.failUnlessEqual(master.checker.users,
                             {"change": "changepw"})
        master.loadConfig(botsCfg)
        self.failUnlessEqual(master.checker.users,
                             {"change": "changepw",
                              "bot1": "pw1",
                              "bot2": "pw2"})
        master.loadConfig(botsCfg)
        self.failUnlessEqual(master.checker.users,
                             {"change": "changepw",
                              "bot1": "pw1",
                              "bot2": "pw2"})
        master.loadConfig(emptyCfg)
        self.failUnlessEqual(master.checker.users,
                             {"change": "changepw"})

    def testSources(self):
        if not cvstoys:
            raise unittest.SkipTest("this test needs CVSToys installed")
        master = self.buildmaster
        master.loadChanges()
        master.loadConfig(emptyCfg)
        self.failUnlessEqual(master.sources, [])
        self.failUnlessEqual(master.change_svc.sources, [])

        master.loadConfig(sourcesCfg)
        self.failUnlessEqual(len(master.sources), 1)
        self.failUnlessEqual(len(master.change_svc.sources), 1)
        s1 = master.change_svc.sources[0]
        self.failUnless(isinstance(s1, FreshCVSSource))
        self.failUnlessEqual(s1.where, ("cvs.example.com", 1000))
        self.failUnlessEqual(s1.prefix, "Prefix/")
        self.failUnlessEqual(s1, master.sources[0])

        # verify that unchanged sources are not interrupted
        master.loadConfig(sourcesCfg)
        self.failUnlessEqual(len(master.change_svc.sources), 1)
        s2 = master.change_svc.sources[0]
        self.failUnlessIdentical(s1, s2)

    def testBuilders(self):
        master = self.buildmaster
        master.loadConfig(emptyCfg)
        self.failUnlessEqual(master.botmaster.builders, {})

        master.loadConfig(buildersCfg)
        self.failUnlessEqual(master.botmaster.builderNames, ["builder1"])
        self.failUnlessEqual(master.botmaster.builders.keys(), ["builder1"])
        b = master.botmaster.builders["builder1"]
        self.failUnless(isinstance(b, Builder))
        self.failUnlessEqual(b.name, "builder1")
        self.failUnlessEqual(b.slavename, "bot1")
        self.failUnlessEqual(b.builddir, "workdir")
        f1 = b.buildFactory
        self.failUnless(isinstance(f1, BasicBuildFactory))
        steps = f1.steps
        self.failUnlessEqual(len(steps), 3)
        self.failUnlessEqual(steps[0], (step.CVS,
                                        {'cvsroot': 'cvsroot',
                                         'cvsmodule': 'cvsmodule',
                                         'mode': 'clobber'}))
        self.failUnlessEqual(steps[1], (step.Compile,
                                        {'command': 'make all'}))
        self.failUnlessEqual(steps[2], (step.Test,
                                        {'command': 'make check'}))


        # make sure a reload of the same data doesn't interrupt the Builder
        master.loadConfig(buildersCfg)
        self.failUnlessEqual(master.botmaster.builderNames, ["builder1"])
        self.failUnlessEqual(master.botmaster.builders.keys(), ["builder1"])
        b2 = master.botmaster.builders["builder1"]
        self.failUnlessIdentical(b, b2)
        # TODO: test that the BuilderStatus object doesn't change
        #statusbag2 = master.client_svc.statusbags["builder1"]
        #self.failUnlessIdentical(statusbag, statusbag2)

        # but changing something should result in a new Builder
        master.loadConfig(buildersCfg2)
        self.failUnlessEqual(master.botmaster.builderNames, ["builder1"])
        self.failUnlessEqual(master.botmaster.builders.keys(), ["builder1"])
        b3 = master.botmaster.builders["builder1"]
        self.failIf(b is b3)
        # the statusbag remains the same TODO
        #statusbag3 = master.client_svc.statusbags["builder1"]
        #self.failUnlessIdentical(statusbag, statusbag3)

        # moving to a new-style builder spec shouldn't cause a change
        master.loadConfig(buildersCfg2new)
        b3n = master.botmaster.builders["builder1"]
        self.failUnlessIdentical(b3n, b3)
        # TODO
        #statusbag3n = master.client_svc.statusbags["builder1"]
        #self.failUnlessIdentical(statusbag3n, statusbag3)

        # unless it is different somehow
        master.loadConfig(buildersCfg1new)
        b3nn = master.botmaster.builders["builder1"]
        self.failIf(b3nn is b3n)

        master.loadConfig(buildersCfg2new)
        b3 = master.botmaster.builders["builder1"]

        # adding new builder
        master.loadConfig(buildersCfg3)
        self.failUnlessEqual(master.botmaster.builderNames, ["builder1",
                                                             "builder2"])
        self.failUnlessListsEquivalent(master.botmaster.builders.keys(),
                                       ["builder1", "builder2"])
        b4 = master.botmaster.builders["builder1"]
        self.failUnlessIdentical(b3, b4)

        # changing first builder should leave it at the same place in the list
        master.loadConfig(buildersCfg4)
        self.failUnlessEqual(master.botmaster.builderNames, ["builder1",
                                                             "builder2"])
        self.failUnlessListsEquivalent(master.botmaster.builders.keys(),
                                       ["builder1", "builder2"])
        b5 = master.botmaster.builders["builder1"]
        self.failIf(b4 is b5)
        
        # and removing it should make the Builder go away
        master.loadConfig(emptyCfg)
        self.failUnlessEqual(master.botmaster.builderNames, [])
        self.failUnlessEqual(master.botmaster.builders, {})
        #self.failUnlessEqual(master.client_svc.statusbags, {}) # TODO

    def checkIRC(self, m, expected):
        ircs = {}
        for irc in self.servers(m, words.IRC):
            ircs[irc.host] = (irc.nick, irc.channels)
        self.failUnlessEqual(ircs, expected)

    def testIRC(self):
        master = self.buildmaster
        master.loadChanges()
        master.loadConfig(emptyCfg)
        self.checkIRC(master, {})

        master.loadConfig(ircCfg1)
        expected = {'irc.us.freenode.net': ('buildbot', ['twisted'])}
        self.checkIRC(master, expected)

        master.loadConfig(ircCfg2)
        expected = {'irc.us.freenode.net': ('buildbot', ['twisted'])}
        self.checkIRC(master, expected)

        master.loadConfig(ircCfg3)
        expected = {'irc.us.freenode.net': ('buildbot', ['twisted']),
                    'irc.example.com': ('otherbot', ['chan1', 'chan2'])}
        self.checkIRC(master, expected)

        master.loadConfig(ircCfg4)
        expected = {'irc.us.freenode.net': ('buildbot', ['knotted'])}
        self.checkIRC(master, expected)

        master.loadConfig(ircCfg1)
        expected = {'irc.us.freenode.net': ('buildbot', ['twisted'])}
        self.checkIRC(master, expected)

    def testWebPortnum(self):
        master = self.buildmaster
        master.loadChanges()

        master.loadConfig(webCfg1)
        ports = self.checkPorts(master, [(9999, pb.PBServerFactory),
                                         (9980, Site)])
        p = ports[1]

        master.loadConfig(webCfg1) # nothing should be changed
        ports = self.checkPorts(master, [(9999, pb.PBServerFactory),
                                         (9980, Site)])
        self.failUnlessIdentical(p, ports[1],
                                 "web port was changed even though " + \
                                 "configuration was not")

        master.loadConfig(webCfg2) # still nothing changes
        ports = self.checkPorts(master, [(9999, pb.PBServerFactory),
                                         (9980, Site)])
        self.failUnlessIdentical(p, ports[1],
                                 "web port was changed even though " + \
                                 "configuration was not")

        master.loadConfig(webCfg3) # changes to 9981
        ports = self.checkPorts(master, [(9999, pb.PBServerFactory),
                                         (9981, Site)])
        self.failIf(p is ports[1],
                    "web port was unchanged but configuration was changed")

        master.loadConfig(emptyCfg)
        self.checkPorts(master, [(9999, pb.PBServerFactory)])

    def testWebPathname(self):
        master = self.buildmaster
        master.loadChanges()

        master.loadConfig(webNameCfg1)
        self.checkPorts(master, [(9999, pb.PBServerFactory),
                                 ('~/.twistd-web-pb', pb.PBServerFactory)])
        unixports = self.UNIXports(master)
        f = unixports[0].args[1]
        self.failUnless(isinstance(f.root, ResourcePublisher))

        master.loadConfig(webNameCfg1) # nothing should be changed
        self.checkPorts(master, [(9999, pb.PBServerFactory),
                                 ('~/.twistd-web-pb', pb.PBServerFactory)])
        self.failUnlessIdentical(f, self.UNIXports(master)[0].args[1],
                                 "web factory was changed even though " + \
                                 "configuration was not")

        master.loadConfig(webNameCfg2) # still nothing changes
        self.checkPorts(master, [(9999, pb.PBServerFactory),
                                 ('~/.twistd-web-pb', pb.PBServerFactory)])
        self.failUnlessIdentical(f, self.UNIXports(master)[0].args[1],
                                 "web factory was changed even though " + \
                                 "configuration was not")

        master.loadConfig(webNameCfg3)
        self.checkPorts(master, [(9999, pb.PBServerFactory),
                                 ('bar.socket', pb.PBServerFactory)])
        self.failIf(f is self.UNIXports(master)[0].args[1],
                    "web factory was unchanged but configuration was changed")

        master.loadConfig(emptyCfg)
        self.checkPorts(master, [(9999, pb.PBServerFactory)])

    def testDebugPassword(self):
        master = self.buildmaster

        master.loadConfig(debugPasswordCfg)
        self.failUnlessEqual(master.checker.users,
                             {"change": "changepw",
                              "debug": "sekrit"})

        master.loadConfig(debugPasswordCfg)
        self.failUnlessEqual(master.checker.users,
                             {"change": "changepw",
                              "debug": "sekrit"})

        master.loadConfig(emptyCfg)
        self.failUnlessEqual(master.checker.users,
                             {"change": "changepw"})

    def checkInterlocks(self, botmaster, expected):
        for (bname, (feeders, interlocks)) in expected.items():
            b = botmaster.builders[bname]
            self.failUnlessListsEquivalent(b.feeders, feeders)
            self.failUnlessListsEquivalent(b.interlocks, interlocks)
        for bname, b in botmaster.builders.items():
            if bname not in expected.keys():
                self.failUnlessEqual(b.feeders, [])
                self.failUnlessEqual(b.interlocks, [])

    def testInterlocks(self):
        master = self.buildmaster
        botmaster = master.botmaster

        # create an inactive interlock
        master.loadConfig(interlockCfg1)
        self.failUnlessListsEquivalent(botmaster.interlocks.keys(),
                                       ['lock1'])
        i1 = botmaster.interlocks['lock1']
        self.failUnless(isinstance(i1, Interlock))
        self.failUnlessEqual(i1.name, 'lock1')
        self.failUnlessEqual(i1.feederNames, ['builder1'])
        self.failUnlessEqual(i1.watcherNames, ['builder2', 'builder3'])
        self.failUnlessEqual(i1.active, False)
        self.checkInterlocks(botmaster, {'builder1': ([], [])})

        # make it active by adding the builder
        master.loadConfig(interlockCfg2)
        self.failUnlessListsEquivalent(botmaster.interlocks.keys(),
                                       ['lock1'])
        # should be the same Interlock object as before
        self.failUnlessIdentical(i1, botmaster.interlocks['lock1'])
        self.failUnless(isinstance(i1, Interlock))
        self.failUnlessEqual(i1.name, 'lock1')
        self.failUnlessEqual(i1.feederNames, ['builder1'])
        self.failUnlessEqual(i1.watcherNames, ['builder2', 'builder3'])
        self.failUnlessEqual(i1.active, True)
        self.checkInterlocks(botmaster, {'builder1': ([i1], []),
                                         'builder2': ([], [i1]),
                                         'builder3': ([], [i1]),
                                         })

        # add a second lock
        master.loadConfig(interlockCfg3)
        self.failUnlessListsEquivalent(botmaster.interlocks.keys(),
                                       ['lock1', 'lock2'])
        self.failUnlessIdentical(i1, botmaster.interlocks['lock1'])
        self.failUnless(isinstance(i1, Interlock))
        self.failUnlessEqual(i1.name, 'lock1')
        self.failUnlessEqual(i1.feederNames, ['builder1'])
        self.failUnlessEqual(i1.watcherNames, ['builder2', 'builder3'])
        self.failUnlessEqual(i1.active, True)
        i2 = botmaster.interlocks['lock2']
        self.failUnless(isinstance(i2, Interlock))
        self.failUnlessEqual(i2.name, 'lock2')
        self.failUnlessEqual(i2.feederNames, ['builder3', 'builder4'])
        self.failUnlessEqual(i2.watcherNames, ['builder5'])
        self.failUnlessEqual(i2.active, True)
        self.checkInterlocks(botmaster, {'builder1': ([i1], []),
                                         'builder2': ([], [i1]),
                                         'builder3': ([i2], [i1]),
                                         'builder4': ([i2], []),
                                         'builder5': ([], [i2]),
                                         })

        # modify the second interlock
        master.loadConfig(interlockCfg4)
        self.failUnlessListsEquivalent(botmaster.interlocks.keys(),
                                       ['lock1', 'lock2'])
        self.failUnlessIdentical(i1, botmaster.interlocks['lock1'])
        self.failUnless(isinstance(i1, Interlock))
        self.failUnlessEqual(i1.name, 'lock1')
        self.failUnlessEqual(i1.feederNames, ['builder1'])
        self.failUnlessEqual(i1.watcherNames, ['builder2', 'builder3'])
        self.failUnlessEqual(i1.active, True)
        # second interlock has changed, should be a new Interlock object
        self.failIf(i2 is botmaster.interlocks['lock2'])
        i2 = botmaster.interlocks['lock2']
        self.failUnless(isinstance(i2, Interlock))
        self.failUnlessEqual(i2.name, 'lock2')
        self.failUnlessEqual(i2.feederNames, ['builder1', 'builder4'])
        self.failUnlessEqual(i2.watcherNames, ['builder5'])
        self.failUnlessEqual(i2.active, True)
        self.checkInterlocks(botmaster, {'builder1': ([i1,i2], []),
                                         'builder2': ([], [i1]),
                                         'builder3': ([], [i1]),
                                         'builder4': ([i2], []),
                                         'builder5': ([], [i2]),
                                         })

        # delete the first interlock
        master.loadConfig(interlockCfg5)
        self.failUnlessEqual(botmaster.interlocks.keys(), ['lock2'])
        self.failUnlessIdentical(i2, botmaster.interlocks['lock2'])
        self.failUnless(isinstance(i2, Interlock))
        self.failUnlessEqual(i2.name, 'lock2')
        self.failUnlessEqual(i2.feederNames, ['builder1', 'builder4'])
        self.failUnlessEqual(i2.watcherNames, ['builder5'])
        self.failUnlessEqual(i2.active, True)
        self.checkInterlocks(botmaster, {'builder1': ([i2], []),
                                         'builder4': ([i2], []),
                                         'builder5': ([], [i2]),
                                         })

        # make it inactive by removing a builder it depends upon
        master.loadConfig(interlockCfg6)
        self.failUnlessEqual(botmaster.interlocks.keys(), ['lock2'])
        self.failUnlessIdentical(i2, botmaster.interlocks['lock2'])
        self.failUnlessEqual(i2.active, False)
        self.checkInterlocks(botmaster, {})

        # now remove it
        master.loadConfig(interlockCfg7)
        self.failUnlessEqual(botmaster.interlocks, {})
        self.checkInterlocks(botmaster, {})


class ConfigFileTest(unittest.TestCase):

    def testFindConfigFile(self):
        os.mkdir("test_cf")
        open(os.path.join("test_cf", "master.cfg"), "w").write(emptyCfg)
        open(os.path.join("test_cf", "alternate.cfg"), "w").write(slaveportCfg)

        m = BuildMaster("test_cf")
        m.loadTheConfigFile()
        self.failUnlessEqual(m.slavePortnum, 9999)

        m = BuildMaster("test_cf", "alternate.cfg")
        m.loadTheConfigFile()
        self.failUnlessEqual(m.slavePortnum, 9000)
