#! /usr/bin/python

from twisted.trial import unittest
from buildbot.process.interlock import Interlock

class FakeWatcher:
    pass

class TestTwoFeeders(unittest.TestCase):

    def setUp(self):
        watchers = []
        self.i = Interlock("i1", ["source1", "source2"], watchers)
        #self.w = FakeWatcher()
        self.doBuild = None
    def callback(self, doBuild):
        self.doBuild = doBuild
    def doSequence(self, actions):
        for (source,num,passed,doBuild) in actions:
            self.i.buildFinished(source, num, passed)
            self.failUnless(self.doBuild == doBuild,
                            "%s,%d,%d !-> %s, not %s" % \
                            (source, num, passed, self.doBuild, doBuild))
    def test1(self):
        # feeders haven't passed any builds
        d = self.i.hasPassed(3)
        d.addCallback(self.callback)
        self.failUnless(self.doBuild == None)
        seq = [("source1", 2, 1, None),
               ("source2", 3, 1, None),
               ("source1", 3, 1, 1),
               ]
        self.doSequence(seq)
    def test2(self):
        d = self.i.hasPassed(3)
        d.addCallback(self.callback)
        self.failUnless(self.doBuild == None)
        seq = [("source1", 2, 1, None),
               ("source2", 3, 0, 0),
               ("source1", 3, 1, 0),
               ]
        self.doSequence(seq)
    def test3(self):
        d = self.i.hasPassed(3)
        d.addCallback(self.callback)
        self.failUnless(self.doBuild == None)
        seq = [("source1", 2, 0, None),
               ("source2", 3, 0, 0),
               ]
        self.doSequence(seq)
    def test4(self):
        self.i.buildFinished("source1", 3, 1)
        self.i.buildFinished("source2", 3, 1)
        d = self.i.hasPassed(3)
        d.addCallback(self.callback)
        self.failUnless(self.doBuild == 1)

class TestOneFeeder(unittest.TestCase):

    def setUp(self):
        watchers = []
        self.i = Interlock("i1", ["source1"], watchers)
        #self.w = FakeWatcher()
        self.doBuild = None
    def callback(self, doBuild):
        self.doBuild = doBuild
    def doSequence(self, actions):
        for (source,num,passed,doBuild) in actions:
            self.i.buildFinished(source, num, passed)
            self.failUnless(self.doBuild == doBuild,
                            "%s,%d,%d !-> %s, not %s" % \
                            (source, num, passed, self.doBuild, doBuild))
    def test1(self):
        d = self.i.hasPassed(3)
        d.addCallback(self.callback)
        self.failUnless(self.doBuild == None)
        self.i.buildFinished("source1", 2, 0)
        self.failUnless(self.doBuild == None)
    def test2(self):
        d = self.i.hasPassed(3)
        d.addCallback(self.callback)
        self.failUnless(self.doBuild == None)
        self.i.buildFinished("source1", 2, 1)
        self.failUnless(self.doBuild == None)
        self.i.buildFinished("source1", 3, 1)
        self.failUnless(self.doBuild == 1)
    def test3(self):
        d = self.i.hasPassed(3)
        d.addCallback(self.callback)
        self.failUnless(self.doBuild == None)
        self.i.buildFinished("source1", 2, 1)
        self.failUnless(self.doBuild == None)
        self.i.buildFinished("source1", 3, 0)
        self.failUnless(self.doBuild == 0)
    def test4(self):
        d = self.i.hasPassed(3)
        d.addCallback(self.callback)
        self.failUnless(self.doBuild == None)
        self.i.buildFinished("source1", 2, 1)
        self.failUnless(self.doBuild == None)
        self.i.buildFinished("source1", 4, 1)
        self.failUnless(self.doBuild == 0)

    def test5(self):
        self.i.buildFinished("source1", 3, 1)
        d = self.i.hasPassed(3)
        d.addCallback(self.callback)
        self.failUnless(self.doBuild == 1)
    def test6(self):
        self.i.buildFinished("source1", 2, 0)
        d = self.i.hasPassed(3)
        d.addCallback(self.callback)
        self.failUnless(self.doBuild == None)
    def test7(self):
        self.i.buildFinished("source1", 2, 1)
        d = self.i.hasPassed(3)
        d.addCallback(self.callback)
        self.failUnless(self.doBuild == None)
        self.i.buildFinished("source1", 3, 1)
        self.failUnless(self.doBuild == 1)
