#! /usr/bin/python

from twisted.trial import unittest
dr = unittest.deferredResult
from twisted.internet import reactor, defer
from twisted.python import log
import sys, os, shutil, time
#log.startLogging(sys.stderr)

from buildbot import master, interfaces
from buildbot.util import now
from buildbot.slave import bot
from buildbot.changes.changes import Change
from buildbot.status import builder

config_1 = """
from buildbot.process import factory

c = {}
c['bots'] = [['bot1', 'sekrit']]
c['sources'] = []
c['builders'] = []
f1 = factory.QuickBuildFactory('fakerep', 'cvsmodule', configure=None)
c['builders'].append(('quick', 'bot1', 'quickdir', f1))
c['slavePortnum'] = 0
BuildmasterConfig = c
"""

config_2 = """
from buildbot.process import factory, step

def s(klass, **kwargs):
    return (klass, kwargs)

f1 = factory.BuildFactory([
    s(step.Dummy, timeout=1),
    s(step.RemoteDummy, timeout=2),
    ])
c = {}
c['bots'] = [['bot1', 'sekrit']]
c['sources'] = []
c['builders'] = [{'name': 'dummy', 'slavename': 'bot1',
                  'builddir': 'dummy', 'factory': f1}]
c['slavePortnum'] = 0
BuildmasterConfig = c
"""

class MyBot(bot.Bot):
    def remote_getSlaveInfo(self):
        return self.parent.info
class MyBuildSlave(bot.BuildSlave):
    botClass = MyBot

class STarget:
    __implements__ = interfaces.IStatusReceiver,
    debug = False

    def __init__(self, mode):
        self.mode = mode
        self.events = []
    def announce(self):
        if self.debug:
            print self.events[-1]

    def builderAdded(self, name, builder):
        self.events.append(("builderAdded", name, builder))
        self.announce()
        if "builder" in self.mode:
            return self
    def builderChangedState(self, name, state, eta):
        self.events.append(("builderChangedState", name, state, eta))
        self.announce()
    def buildStarted(self, name, build):
        self.events.append(("buildStarted", name, build))
        self.announce()
        if "eta" in self.mode:
            self.eta_build = build.getETA()
        if "build" in self.mode:
            return self
    def buildETAUpdate(self, build, ETA):
        self.events.append(("buildETAUpdate", build, ETA))
        self.announce()
    def stepStarted(self, build, step):
        self.events.append(("stepStarted", build, step))
        self.announce()
        if 0 and "eta" in self.mode:
            print "TIMES", step.getTimes()
            print "ETA", step.getETA()
            print "EXP", step.getExpectations()
        if "step" in self.mode:
            return self
    def stepETAUpdate(self, build, step, ETA, expectations):
        self.events.append(("stepETAUpdate", build, step, ETA, expectations))
        self.announce()
    def logStarted(self, build, step, log):
        self.events.append(("logStarted", build, step, log))
        self.announce()
    def logFinished(self, build, step, log):
        self.events.append(("logFinished", build, step, log))
        self.announce()
    def stepFinished(self, build, step, results):
        self.events.append(("stepFinished", build, step, results))
        if 0 and "eta" in self.mode:
            print "post-EXP", step.getExpectations()
        self.announce()
    def buildFinished(self, name, build, results):
        self.events.append(("buildFinished", name, build, results))
        self.announce()
    def builderRemoved(self, name):
        self.events.append(("builderRemoved", name))
        self.announce()

class Run(unittest.TestCase):
    def testMaster(self):
        shutil.rmtree("basedir", ignore_errors=1)
        os.mkdir("basedir")
        m = master.BuildMaster("basedir")
        m.loadConfig(config_1)
        m.readConfig = True
        m.startService()
        cm = m.change_svc
        c = Change("bob", ["Makefile", "foo/bar.c"], "changed stuff")
        cm.addChange(c)
        b1 = m.botmaster.builders["quick"]
        self.failUnless(b1.waiting)
        # now kill the timer
        b1.waiting.stopTimer()

class RunMixin:
    master = None
    slave = None
    slave2 = None

    def setUp(self):
        shutil.rmtree("basedir", ignore_errors=1)
        shutil.rmtree("slavebase", ignore_errors=1)
        shutil.rmtree("slavebase2", ignore_errors=1)
        os.mkdir("basedir")
        self.master = master.BuildMaster("basedir")

    def connectSlave(self):
        port = self.master.slavePort._port.getHost().port
        os.mkdir("slavebase")
        slave = MyBuildSlave("localhost", port, "bot1", "sekrit",
                             "slavebase", keepalive=0, usePTY=1)
        slave.info = {"admin": "one"}
        self.slave = slave
        slave.startService()
        d = self.master.botmaster.waitUntilBuilderAttached("dummy")
        dr(d)

    def connectSlave2(self):
        port = self.master.slavePort._port.getHost().port
        os.mkdir("slavebase2")
        slave = MyBuildSlave("localhost", port, "bot1", "sekrit",
                             "slavebase2", keepalive=0, usePTY=1)
        slave.info = {"admin": "two"}
        self.slave2 = slave
        slave.startService()

    def connectSlave3(self):
        # this slave has a very fast keepalive timeout
        port = self.master.slavePort._port.getHost().port
        os.mkdir("slavebase")
        slave = MyBuildSlave("localhost", port, "bot1", "sekrit",
                             "slavebase", keepalive=2, usePTY=1,
                             keepaliveTimeout=1)
        slave.info = {"admin": "one"}
        self.slave = slave
        slave.startService()
        d = self.master.botmaster.waitUntilBuilderAttached("dummy")
        dr(d)

    def tearDown(self):
        log.msg("doing tearDown")
        self.shutdownSlave()
        if self.master:
            dr(defer.maybeDeferred(self.master.stopService))
            self.master = None
        log.msg("tearDown done")

    # various forms of slave death

    def shutdownSlave(self, waitForMasterToo=True):
        # the slave has disconnected normally: they SIGINT'ed it, or it shut
        # down willingly. This will kill child processes and give them a
        # chance to finish up.
        if self.slave:
            d = self.slave.waitUntilDisconnected()
            dr(defer.maybeDeferred(self.slave.stopService))
            dr(d)
            self.slave = None
        if self.slave2:
            d = self.slave2.waitUntilDisconnected()
            dr(defer.maybeDeferred(self.slave2.stopService))
            dr(d)
            self.slave2 = None
        if waitForMasterToo:
            d = self.master.botmaster.waitUntilBuilderDetached("dummy")
            dr(d)

    def killSlave(self):
        # the slave has died, its host sent a FIN. The .notifyOnDisconnect
        # callbacks will terminate the current step, so the build should be
        # flunked (no further steps should be started).
        self.slave.bf.continueTrying = 0
        bot = self.slave.getServiceNamed("bot")
        broker = bot.builders["dummy"].remote.broker
        broker.transport.loseConnection()
        self.slave = None

    def disappearSlave(self):
        # the slave's host has vanished off the net, leaving the connection
        # dangling. This will be detected quickly by app-level keepalives or
        # a ping, or slowly by TCP timeouts.

        # implement this by replacing the slave Broker's .dataReceived method
        # with one that just throws away all data.
        def discard(data):
            pass
        bot = self.slave.getServiceNamed("bot")
        broker = bot.builders["dummy"].remote.broker
        broker.dataReceived = discard # seal its ears
        broker.transport.write = discard # and take away its voice

    def ghostSlave(self):
        # the slave thinks it has lost the connection, and initiated a
        # reconnect. The master doesn't yet realize it has lost the previous
        # connection, and sees two connections at once.
        raise NotImplementedError

class Status(RunMixin, unittest.TestCase):

    def testSlave(self):
        m = self.master
        s = m.getStatus()
        t1 = STarget(["builder"])
        #t1.debug = True; print
        s.subscribe(t1)
        self.failUnlessEqual(len(t1.events), 0)

        t3 = STarget(["builder", "build", "step"])
        s.subscribe(t3)

        m.loadConfig(config_2)
        m.readConfig = True
        m.startService()

        self.failUnlessEqual(len(t1.events), 2)
        self.failUnlessEqual(t1.events[0][0:2], ("builderAdded", "dummy"))
        self.failUnlessEqual(t1.events[1],
                             ("builderChangedState", "dummy", "offline", None))
        t1.events = []

        self.failUnlessEqual(s.getBuilderNames(), ["dummy"])
        s1 = s.getBuilder("dummy")
        self.failUnlessEqual(s1.getName(), "dummy")
        self.failUnlessEqual(s1.getState(), ("offline", None, None))
        self.failUnlessEqual(s1.getCurrentBuild(), None)
        self.failUnlessEqual(s1.getLastFinishedBuild(), None)
        self.failUnlessEqual(s1.getBuild(-1), None)
        #self.failUnlessEqual(s1.getEvent(-1), foo("created"))

        # status targets should, upon being subscribed, immediately get a
        # list of all current builders
        t2 = STarget([])
        s.subscribe(t2)
        self.failUnlessEqual(len(t2.events), 1)
        self.failUnlessEqual(t2.events[0][0:2], ("builderAdded", "dummy"))

        self.connectSlave()

        self.failUnlessEqual(len(t1.events), 1)
        self.failUnlessEqual(t1.events[0],
                             ("builderChangedState", "dummy", "idle", None))
        t1.events = []

        c = interfaces.IControl(m)
        bc = c.getBuilder("dummy").forceBuild(None,
                                              "forced build for testing")
        d = bc.getStatus().waitUntilFinished()
        res = dr(d)

        # t1 subscribes to builds, but not anything lower-level
        ev = t1.events
        self.failUnlessEqual(len(ev), 4)
        self.failUnlessEqual(ev[0][0:3],
                             ("builderChangedState", "dummy", "building"))
        self.failUnlessEqual(ev[1][0], "buildStarted")
        self.failUnlessEqual(ev[2][0:2]+ev[2][3:4],
                             ("buildFinished", "dummy", builder.SUCCESS))
        self.failUnlessEqual(ev[3],
                             ("builderChangedState", "dummy", "idle", None))

        self.failUnlessEqual([ev[0] for ev in t3.events],
                             ["builderAdded",
                              "builderChangedState", # offline
                              "builderChangedState", # idle
                              "builderChangedState", # building
                              "buildStarted",
                              "stepStarted", "stepETAUpdate", "stepFinished",
                              "stepStarted", "stepETAUpdate",
                              "logStarted", "logFinished", "stepFinished",
                              "buildFinished",
                              "builderChangedState", # idle
                              ])

        b = s1.getLastFinishedBuild()
        self.failUnless(b)
        self.failUnlessEqual(b.getBuilder().getName(), "dummy")
        self.failUnlessEqual(b.getNumber(), 0)
        self.failUnlessEqual(b.getSourceStamp(), (None, None))
        self.failUnlessEqual(b.getReason(), "forced build for testing")
        self.failUnlessEqual(b.getChanges(), [])
        self.failUnlessEqual(b.getResponsibleUsers(), [])
        self.failUnless(b.isFinished())
        self.failUnlessEqual(b.getText(), ['build', 'successful'])
        self.failUnlessEqual(b.getColor(), "green")
        self.failUnlessEqual(b.getResults(), builder.SUCCESS)

        steps = b.getSteps()
        self.failUnlessEqual(len(steps), 2)

        st1 = steps[0]
        self.failUnlessEqual(st1.getName(), "dummy")
        self.failUnless(st1.isFinished())
        self.failUnlessEqual(st1.getText(), ["delay", "1 secs"])
        start,finish = st1.getTimes()
        self.failUnless(0.5 < (finish-start) < 10)
        self.failUnlessEqual(st1.getExpectations(), [])
        self.failUnlessEqual(st1.getLogs(), [])

        st2 = steps[1]
        self.failUnlessEqual(st2.getName(), "remote dummy")
        self.failUnless(st2.isFinished())
        self.failUnlessEqual(st2.getText(),
                             ["remote", "delay", "2 secs"])
        start,finish = st2.getTimes()
        self.failUnless(1.5 < (finish-start) < 10)
        self.failUnlessEqual(st2.getExpectations(), [('output', 38, None)])
        logs = st2.getLogs()
        self.failUnlessEqual(len(logs), 1)
        self.failUnlessEqual(logs[0].getName(), "log")
        self.failUnlessEqual(logs[0].getText(), "data")

        # now we run it a second time, and we should have an ETA

        t4 = STarget(["builder", "build", "eta"])
        s.subscribe(t4)
        c = interfaces.IControl(m)
        bc = c.getBuilder("dummy").forceBuild(None,
                                              "forced build for testing")
        d = bc.getStatus().waitUntilFinished()
        res = dr(d)

        self.failUnless(3.0 < t4.eta_build < 5.0) # should be 4 seconds

class Disconnect(RunMixin, unittest.TestCase):

    def disconnectSetupMaster(self):
        # verify that disconnecting the slave during a build properly
        # terminates the build
        m = self.master
        s = m.getStatus()
        c = interfaces.IControl(m)

        m.loadConfig(config_2)
        m.readConfig = True
        m.startService()

        self.failUnlessEqual(s.getBuilderNames(), ["dummy"])
        s1 = s.getBuilder("dummy")
        self.failUnlessEqual(s1.getName(), "dummy")
        self.failUnlessEqual(s1.getState(), ("offline", None, None))
        self.failUnlessEqual(s1.getCurrentBuild(), None)
        self.failUnlessEqual(s1.getLastFinishedBuild(), None)
        self.failUnlessEqual(s1.getBuild(-1), None)
        return m,s,c,s1

    def disconnectSetup(self):
        m,s,c,s1 = self.disconnectSetupMaster()
        self.connectSlave()
        self.failUnlessEqual(s1.getState(), ("idle", None, None))
        return m,s,c,s1

    def disconnectSetup2(self):
        m,s,c,s1 = self.disconnectSetupMaster()
        self.connectSlave3()
        self.failUnlessEqual(s1.getState(), ("idle", None, None))
        return m,s,c,s1


    def verifyDisconnect(self, bs):
        self.failUnless(bs.isFinished())

        step1 = bs.getSteps()[0]
        self.failUnlessEqual(step1.getText(), ["delay", "interrupted"])
        self.failUnlessEqual(step1.getResults()[0], builder.FAILURE)

        self.failUnlessEqual(bs.getResults(), builder.FAILURE)

    def verifyDisconnect2(self, bs):
        self.failUnless(bs.isFinished())

        step1 = bs.getSteps()[1]
        self.failUnlessEqual(step1.getText(), ["remote", "delay", "2 secs",
                                               "failed", "slave", "lost"])
        self.failUnlessEqual(step1.getResults()[0], builder.FAILURE)

        self.failUnlessEqual(bs.getResults(), builder.FAILURE)


    def testIdle1(self):
        m,s,c,s1 = self.disconnectSetup()
        # disconnect the slave before the build starts
        self.shutdownSlave() # dies before it gets started

        # trying to force a build now will cause an error. Regular builds
        # just wait for the slave to re-appear, but forced builds that
        # cannot be run right away trigger NoSlaveErrors
        fb = c.getBuilder("dummy").forceBuild
        self.failUnlessRaises(interfaces.NoSlaveError,
                              fb, None, "forced build")

    def testIdle2(self):
        m,s,c,s1 = self.disconnectSetup()
        # now suppose the slave goes missing
        self.disappearSlave()

        # forcing a build will work: the build will begin, since we think we
        # have a slave. The build will fail, however, because of a timeout
        # error.
        bc = c.getBuilder("dummy").forceBuild(None, "forced build")
        bs = bc.getStatus()
        print "build started"
        d = bs.waitUntilFinished()
        dr(d, 5)
        print bs.getText()
    testIdle2.skip = "short timeout not yet implemented"

    def testSlaveTimeout(self):
        m,s,c,s1 = self.disconnectSetup2() # fast timeout

        # now suppose the slave goes missing. We want to find out when it
        # creates a new Broker, so we reach inside and mark it with the
        # well-known sigil of impending messy death.
        bd = self.slave.getServiceNamed("bot").builders["dummy"]
        broker = bd.remote.broker
        broker.redshirt = 1

        # make sure the keepalives will keep the connection up
        later = now() + 5
        while 1:
            if now() > later:
                break
            bd = self.slave.getServiceNamed("bot").builders["dummy"]
            if not bd.remote or not hasattr(bd.remote.broker, "redshirt"):
                self.fail("slave disconnected when it shouldn't have")
            reactor.iterate(0.01)

        d = self.master.botmaster.waitUntilBuilderDetached("dummy")
        # whoops! how careless of me.
        self.disappearSlave()

        # the slave will realize the connection is lost within 2 seconds, and
        # reconnect.
        dr(d, 5)
        d = self.master.botmaster.waitUntilBuilderAttached("dummy")
        dr(d, 5)
        # make sure it is a new connection (i.e. a new Broker)
        bd = self.slave.getServiceNamed("bot").builders["dummy"]
        self.failUnless(bd.remote, "hey, slave isn't really connected")
        self.failIf(hasattr(bd.remote.broker, "redshirt"),
                    "hey, slave's Broker is still marked for death")

    def testBuild1(self):
        m,s,c,s1 = self.disconnectSetup()
        # this next sequence is timing-dependent. The dummy build takes at
        # least 3 seconds to complete, and this batch of commands must
        # complete within that time.
        #
        bc = c.getBuilder("dummy").forceBuild(None, "forced build")
        bs = bc.getStatus()
        # kill the slave while it's running the first step
        self.shutdownSlave() # dies before it gets started

        # now examine the just-stopped build and make sure it is really
        # stopped. This is checking for bugs in which the slave-detach gets
        # missed or causes an exception which prevents the build from being
        # marked as "finished due to an error".
        d = bs.waitUntilFinished()
        dr(d, 5)

        self.failUnlessEqual(s1.getState()[0], "offline")
        self.verifyDisconnect(bs)

    def testBuild2(self):
        m,s,c,s1 = self.disconnectSetup()
        # this next sequence is timing-dependent
        bc = c.getBuilder("dummy").forceBuild(None, "forced build")
        bs = bc.getStatus()
        # shutdown the slave while it's running the first step
        reactor.callLater(0.5, self.shutdownSlave)

        dr(bs.waitUntilFinished(), 5)

        self.failUnlessEqual(s1.getState()[0], "offline")
        self.verifyDisconnect(bs)

    def testBuild3(self):
        m,s,c,s1 = self.disconnectSetup()
        # this next sequence is timing-dependent
        bc = c.getBuilder("dummy").forceBuild(None, "forced build")
        bs = bc.getStatus()
        # kill the slave while it's running the first step
        reactor.callLater(0.5, self.killSlave)

        dr(bs.waitUntilFinished(), 5)

        self.failUnlessEqual(s1.getState()[0], "offline")
        self.verifyDisconnect(bs)

    def testBuild4(self):
        m,s,c,s1 = self.disconnectSetup()
        # this next sequence is timing-dependent
        bc = c.getBuilder("dummy").forceBuild(None, "forced build")
        bs = bc.getStatus()
        # kill the slave while it's running the second (remote) step
        reactor.callLater(1.5, self.killSlave)

        dr(bs.waitUntilFinished(), 5)

        self.failUnlessEqual(s1.getState()[0], "offline")
        self.verifyDisconnect2(bs)

    def testInterrupt(self):
        m,s,c,s1 = self.disconnectSetup()
        # this next sequence is timing-dependent
        bc = c.getBuilder("dummy").forceBuild(None, "forced build")
        bs = bc.getStatus()
        # halt the build while it's running the first step
        reactor.callLater(0.5, bc.stopBuild, "bang go splat")

        dr(bs.waitUntilFinished(), 5)

        self.verifyDisconnect(bs)

    def testDisappear(self):
        m,s,c,s1 = self.disconnectSetup()
        bc = c.getBuilder("dummy")

        # ping should succeed
        d = bc.ping(1)
        res = dr(d)
        self.failUnlessEqual(res, True)

        # now, before any build is run, make the slave disappear
        self.slave.bf.continueTrying = 0
        self.disappearSlave()

        # at this point, a ping to the slave should timeout
        d = bc.ping(1)
        res = dr(d)
        self.failUnlessEqual(res, False)

    def testDuplicate(self):
        m,s,c,s1 = self.disconnectSetup()
        bc = c.getBuilder("dummy")
        bs = s.getBuilder("dummy")
        ss = bs.getSlave()

        self.failUnless(ss.isConnected())
        self.failUnlessEqual(ss.getAdmin(), "one")

        # now, before any build is run, make the first slave disappear
        self.slave.bf.continueTrying = 0
        self.disappearSlave()

        d = self.master.botmaster.waitUntilBuilderDetached("dummy")
        # now let the new slave take over
        self.connectSlave2()
        dr(d, 2)
        d = self.master.botmaster.waitUntilBuilderAttached("dummy")
        dr(d, 2)

        self.failUnless(ss.isConnected())
        self.failUnlessEqual(ss.getAdmin(), "two")
