#! /usr/bin/python

import sys, os
from twisted.python import log, components
#log.startLogging(sys.stderr)

from twisted.trial import unittest
dr = unittest.deferredResult

from twisted.internet import reactor
from twisted.internet.interfaces import IReactorUNIX
from twisted.web import client

from buildbot import master, interfaces
from buildbot.status import html

class ConfiguredMaster(master.BuildMaster):
    """This BuildMaster variant has a static config file, provided as a
    string when it is created."""

    def __init__(self, basedir, config):
        self.config = config
        master.BuildMaster.__init__(self, basedir)
        
    def loadTheConfigFile(self):
        self.loadConfig(self.config)

components.registerAdapter(master.Control, ConfiguredMaster,
                           interfaces.IControl)


config1 = """
BuildmasterConfig = {
    'bots': [],
    'sources': [],
    'builders': [],
    'slavePortnum': 0,
    '%(k)s': %(v)s,
    }
"""



def distrib_unix(unixpath):
    from twisted.web import server, resource, distrib
    root = resource.Resource()
    r = distrib.ResourceSubscription("unix", unixpath)
    root.putChild('remote', r)
    p = reactor.listenTCP(0, server.Site(root))
    return p

def distrib_tcp(port):
    from twisted.web import server, resource, distrib
    root = resource.Resource()
    r = distrib.ResourceSubscription("localhost", port)
    root.putChild('remote', r)
    p = reactor.listenTCP(0, server.Site(root))
    return p


class WebTest(unittest.TestCase):
    def tearDown(self):
        # grr.
        from twisted.protocols import http
        http._logDateTimeStop()

    def find_waterfall(self, master):
        return filter(lambda child: isinstance(child, html.Waterfall),
                      list(master))

    def test_webPortnum(self):
        cfg = config1 % {'k': 'webPortnum',
                         'v': "0"}
        os.mkdir("test_web1")
        m = ConfiguredMaster("test_web1", cfg)
        m.startService()
        # hack to find out what randomly-assigned port it is listening on
        port = list(self.find_waterfall(m)[0])[0]._port.getHost().port

        d = client.getPage("http://localhost:%d/" % port)
        page = dr(d)
        #print page
        self.failUnless(page)
        dr(m.stopService())

    def test_webPathname(self):
        # running a t.web.distrib server over a UNIX socket
        if not components.implements(reactor, IReactorUNIX):
            raise unittest.SkipTest("UNIX sockets not supported here")

        cfg = config1 % {'k': 'webPathname',
                         'v': "'.web-pb'"}

        os.mkdir("test_web2")
        m = ConfiguredMaster("test_web2", cfg)
        m.startService()
            
        p = distrib_unix("test_web2/.web-pb")
        port = p.getHost().port

        d = client.getPage("http://localhost:%d/remote/" % port)
        page = dr(d)
        #print page
        self.failUnless(page)
        dr(p.stopListening())
        dr(m.stopService())
        

    def test_webPathname_port(self):
        # running a t.web.distrib server over TCP
        cfg = config1 % {'k': 'webPathname',
                         'v': "0"}

        os.mkdir("test_web3")
        m = ConfiguredMaster("test_web3", cfg)
        m.startService()
        dport = list(self.find_waterfall(m)[0])[0]._port.getHost().port

        p = distrib_tcp(dport)
        port = p.getHost().port

        d = client.getPage("http://localhost:%d/remote/" % port)
        page = dr(d)
        #print page
        self.failUnless(page)
        dr(p.stopListening())
        dr(m.stopService())
        
